// --- [ dog_gfx.c ] --------------------------------------------------------

/*
Dieses Modul ist fr die Ausgabe von Grafik zustndig.
Fr die Verwaltung muss ein Speicherbereich im Controller reserviert werden,
der den Grafikbreich auf dem Display reprsentiert.
Die Grafik wird pixelweise in den Speicher geschrieben,
anschlieend wird der Speicher komplett ins Display bertragen.

Die Routinen sind inspiriert (aber nicht kopiert!) vom Arduino-Projekt von Oliver Kraus
*/

#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <avr/io.h>
#include "global.h"
#include "dog128x64.h"
#include "dog_small_font.h"
#include "dog_gfx.h"

#if DEMO == GRAPHIK_DEMO
#define DOG_GFX
#endif

#ifdef DOG_GFX
// --- externe Funktionen
void DOG_write_data(unsigned char data);
void DOG_write_cmd(unsigned char data);

// --- lokale Funktionen
void DOG_set_pixel(void);
void DOG_clear_pixel(void);
void DOG_xor_pixel(void);


// --- Variablen des Moduls
uint8_t gx, gy, page, col;
uint8_t screen[DOB_PAGE_CNT][DOG_GX_PIXEL];

// Bitmuster fr einen einzelnen Punkt (von oben nach unten)
static uint8_t g_point[8] = {1, 2, 4, 8, 16, 32, 64, 128};
// Bitmuster fr das obere, vertikale Linienende
static uint8_t g_line[8]  = {255, 254, 252, 248, 240, 224, 192, 128};
// das Bitmuster fr das untere, vertikale Linienende wird durch
// Negieren des oberen Linienendes erzeugt !

// Die Adressierung des Displays
// Der Grafikspeicher ist in Zeilen (g_row) organisiert.
// Die oberste Zeile hat die Nr. 0, nach unten nimmt der Wert zu.
// Jede Zeile besteht aus 8 vertikalen Pixeln,
// das oberste Pixel hat den Wert 1, ... das unterste den Wert 128 (-> g_point[]).

// Das Koordinatensystem hat seinen Ursprung (0,0) in der oberen, linken Ecke.
// Die positive Y-Achse zeigt nach unten !!

// Die Adressierung der Grafik:
// Der Koordinatenursprung liegt links unten im Grafikbereich !!!
// Es wird intern umgerechnet auf die Adressierung des Displays.

// ----------------------------------------------------------------------------
// Setzt ein Pixel an der Position, die von den Parametern x uns y angegeben werden
// Die x-Koordinate kann direkt zur Adressierung benutzt werden,
// aus der Y-Koordinate muss der Buffer berechnet werden

void DOG_set_pixel_xy(uint8_t x, uint8_t y)
{
/*
gx = x;
gy = (DOG_GY_PIXEL -1) -y;

page = gy / 8;									   // den Buffer bestimmen
screen[page][gx] |= g_point[gy % 8];      // den Punkt setzen
*/

screen[y/8][x] |= g_point[y % 8];      	// den Punkt setzen


}

// ----------------------------------------------------------------------------
// Lscht ein Pixel an der Position, die von den Parametern x uns y angegeben werden
void DOG_clr_pixel_xy(uint8_t x, uint8_t y)
{
screen[y/8][x] &= ~g_point[y % 8];      // den Punkt lschen
}


// ----------------------------------------------------------------------------
// XOR ein Pixel an der Position, die von den Parametern x uns y angegeben werden
void DOG_xor_pixel_xy(uint8_t x, uint8_t y)
{
screen[y/8][x] ^= g_point[y % 8];      // den Punkt XORdieren
}


// ----------------------------------------------------------------------------
// Zeichnet eine Linie zwischen beliebigen Punkten
// Zuerst wird die Neigung der Achse geprft.
// Entlang der flacheren Achse hangelt sich das Programm von Pixel zu Pixel und
// bestimmt die jeweils zugehrige 2. Pixelkoordinate.
// Die Achsen werden jeweils in aufsteigender Richtung abgefahren.
// Die Routine zur Berechnung der Linienpixel stammt von Oliver Kraus
// und ist der Arduino library entnommen.

void DOG_line(uint8_t x1, uint8_t y1, uint8_t x2, uint8_t y2, uint8_t mode)
{
int8_t delta_x, delta_y, step, i;

y1 = (DOG_GY_PIXEL-1) - y1;
y2 = (DOG_GY_PIXEL-1) - y2;

if (y1 == y2)										// eine horizontale Linie
	{
	DOG_hline(x1, x2, y1, mode);
	return;
	}
if (x1 == x2)										// eine vertikale Linie
	{
	DOG_vline(x1, y1, y2, mode);
	return;
	}

delta_x = abs(x2 - x1);							// Neigungsverhltnis der Linie bestimmen
delta_y = abs(y2 - y1);
														// wenn delta_x >= delta_y, dann ist die
														// Linie max. mit 45 gegen die X-Achse geneigt,
if (delta_x >= delta_y)							// dann entlang der X-Achse abtasten
	{
	if (x1 > x2)									// ggf. Anfangs-/EndPunkt tauschen
		{
		tmp = x1;
		x1 = x2;
		x2 = tmp;
		tmp = y1;
		y1 = y2;
		y2 = tmp;
		}

	delta_x = x2 - x1;                    // nun ist x2 definitiv grer als x1
	delta_y = y2 - y1;                    //

   if (delta_y > 0) step = 1; else step = -1;
   tmp = delta_x / 2;

	for( i = x1; i <= x2; i++)
		{
      if (mode == SET_PIX) 		DOG_set_pixel_xy(i, y1);
		else if (mode == CLR_PIX) 	DOG_set_pixel_xy(i, y1);
		else 								DOG_xor_pixel_xy(i, y1);

		tmp -= abs(delta_y);

      if ((int8_t)tmp < 0)
         {
         tmp += delta_x;
         y1 += step;
			}
      }
	}

else													// dann entlang der Y-Achse abtasten
	{
	if (y1 > y2)									// ggf. Anfangs-/EndPunkt tauschen
		{
		tmp = x1;
		x1 = x2;
		x2 = tmp;
		tmp = y1;
		y1 = y2;
		y2 = tmp;
		}
   delta_x = x2 - x1;
	delta_y = y2 - y1;

   if (delta_x > 0) step = 1; else step = -1;

   tmp = delta_y / 2;

	for( i = y1; i <= y2; i++)
		{
      if (mode == SET_PIX)			DOG_set_pixel_xy(x1, i);
		else if (mode == CLR_PIX)	DOG_clr_pixel_xy(x1, i);
		else 								DOG_xor_pixel_xy(x1, i);

		tmp -= abs(delta_x);

      if ((int8_t)tmp < 0)
         {
         tmp += delta_y;
			x1 += step;
			}
      }
   }
}

// ----------------------------------------------------------------------------
// Sonderfall: zeichnet ein horizontale Linie
// hier muss nur 1-mal die Bitmaske fr das gewnschte Pixel ermittelt und
// anschlieend entlang der X-Achse wiederholt gesetzt werden.
void DOG_hline(uint8_t x1, uint8_t x2, uint8_t y, uint8_t mode)
{
uint8_t bit_mask;
gx = x1;
y = (DOG_GY_PIXEL - 1) - y;

page = y/8;
bit_mask = g_point[y % 8];
if (mode == SET_PIX)
	{
	for (tmp = x1; tmp <= x2; tmp++) screen[page][tmp] |= bit_mask;
	}
else if (mode == CLR_PIX)
	{
	for (tmp = x1; tmp <= x2; tmp++) screen[page][tmp] &= ~bit_mask;
	}
else
	{
	for (tmp = x1; tmp <= x2; tmp++) screen[page][tmp] ^= bit_mask;
	}

}

// ----------------------------------------------------------------------------
// Sonderfall: zeichnet ein vertikale Linie
// die Ausgabe erfolgt zeilenweise mit jeweils einem Byte als Bitmap der Linie.
// Es gibt drei Mglichkeiten:
// 1. Die Linie durchquert den Zeile komplett:
//    Das Bitmap der Linie entspricht 255
// 2. Die Linie hat die oberste Zeile erreicht:
//    Das Bitmap der Linie wird aus g_line[] geholt.
// 3. Die Linie hat die unterste Zeile erreicht:
//    Das Bitmap der Linie wird aus g_line[] geholt und negiert

void DOG_vline(uint8_t x, uint8_t y1, uint8_t y2, uint8_t mode)
{
uint8_t oben, unten;

y1 = (DOG_GY_PIXEL - 1) - y1;
y2 = (DOG_GY_PIXEL - 1) - y2;

if (y1 > y2)						// y1 muss oberhalb von y2 liegen
	{									// im Sinne der Pagezhlung !!
	tmp = y1;						// nicht bezogen auf das Koordinatensystem
	y1 = y2;
	y2 = tmp;
	}

oben  = y1 / 8;
unten = y2 / 8;

for (page = oben; page <= unten; page++)
	{
	if (page == oben)
		{
		tmp = g_line[y1 % 8];
		}

	else if (page == unten)
		{
		tmp = ~(g_line[y2 % 8]);
		}

	else
		{
		tmp = 255;
		}

	if (mode == SET_PIX) 		screen[page][x] |= tmp;
	else if (mode == CLR_PIX) 	screen[page][x] &= ~tmp;
	else								screen[page][x] ^= tmp;
	}
}

// ----------------------------------------------------------------------------
// zeichnet einen Rahmen aus 4 einzelnen Linien
void DOG_box(uint8_t x1, uint8_t y1, uint8_t x2, uint8_t y2, uint8_t mode)
{
y1 = (DOG_GY_PIXEL-1) - y1;
y2 = (DOG_GY_PIXEL-1) - y2;

DOG_hline(x1, x2, y1, mode);
DOG_hline(x1, x2, y2, mode);

// bei einem XOR drfen die Eckpunkte nicht zweifach XOR werden,
// daher wird die vertikale Linie beidseitig um 1 Punkt gekrzt

if (y1 < y2)
	{
	y2--;
	}

else
	{
	y1--;
	}

DOG_vline(x1, y1, y2, mode);
DOG_vline(x2, y1, y2, mode);
}

// ----------------------------------------------------------------------------
// Sendet den Grafikspeicher an das Display
// nach dem Zeichnen von Linien etc. steht die Grafik zunchst nur im Speicher.
// Erst ein DOG_write_gfx() bertrgt den Speicher auf das Display und aktualisiert die Anzeige !
void DOG_write_gfx(void)
{
for (page = 0; page < DOB_PAGE_CNT; page++)
	{
	DOG_goto_xy(DOG_GX_OFFSET, page);
	for (col = 0; col < DOG_GX_PIXEL; col++) DOG_write_data(screen[page][col]);
	}
}

// ----------------------------------------------------------------------------
// Lscht nur den Grafikspeicher, nicht aber das Display
// soll auch das Display gelscht werden, dann ist ein DOG_write_gfx() erforderlich !!
void DOG_clear_gfx(void)
{
for (page = 0; page < DOB_PAGE_CNT; page++)
	{
	for (col = 0; col < DOG_GX_PIXEL; col++) screen[page][col] = 0;
	}
}

// ----------------------------------------------------------------------------
// Setzt die Variable gx/gy fr den Startpunkt eines Textes im Grafikmodus:
// der Startpunkt ist unabhngig von der Pagestruktur und kann in X- und Y-Richtung
// frei gewhlt werden.
// Der Ursprung 0/0 liegt unten links !!
// Rechnet dabei auf das Display-interne Koordinatensystem um

void DOG_set_gfx_position(uint8_t x, uint8_t y)
{
gx = x;
gy = (DOG_GY_PIXEL - 1) - y;
}

// ----------------------------------------------------------------------------
// Gibt im Grafikmodus einen String (als Ziffernfolge) aus.
// Die ASCII-Werte werden als Index (Zahl) benutzt,
// der Zeichensatz ist hier nur fr die 10 Ziffern + Dezimalpunkt definiert !!!!
// -> siehe DOG_small_font.h
void DOG_print_gs(char * str, uint8_t mode)
{
while (*str) DOG_print_gc((*str++) - '0', mode);
}

// ----------------------------------------------------------------------------
// Gibt einen Text (hier nur Ziffern) bis 8 Pixel Hhe im Grafikmodus aus,
// schreibt die Pixel in den pageorientierten Grafikspeicher.
//
// Die Startposition des Textes wird von den Variablen gx/gy bestimmt.
// Die Variable gx wird um die Textbreite incrementiert, so dass mehrere Zeichen
// nacheinander geschrieben werden knnen, ohne dass die Textposition angepasst werden muss.
//
// Die im Hintergrund stehende Page-Struktur muss bercksichtigt werden:
// je nach der Startposition und der Fonthhe kann das Zeichen
// - innerhalb einer Page liegen
// - sich ber zwei Pages erstrecken
// Fr Schrifthhen grer als 8 Pixel ist diese Funktion nicht ausgelegt !!!
// Denn hier knnte das Zeichen sich auch ber drei oder mehr Pages ausdehnen.
//
// Diese Funktion ist fr den small-font ausgelegt - fr andere Fonts sind Anpassungen notwendig.

void DOG_print_gc(uint8_t idx, uint8_t mode)
{
uint8_t i, pos_y;
char * ptr;                               // der Pointer auf die Bytes des Zeichens

uint8_t upper_pix[SMALL_FONT_WIDTH];      // zur Aufnahme/Shiften der Zeichenbits
uint8_t lower_pix[SMALL_FONT_WIDTH];      // zur Aufnahme/Shiften der Zeichenbits
if (idx > 10) idx = 10;                   // es sind z.Z. nur die 10 Ziffern und der Dezimalpunkt definiert !

ptr = &small_font[idx][0];       			// den Pointer auf das gesuchte Zeichen setzen
for (i = 0; i < SMALL_FONT_WIDTH; i++)    // die Bitfolgen einlesen
   {
   lower_pix[i] = pgm_read_byte(ptr++);	// ins lower-Byte laden
   upper_pix[i] = 0;
	}

page = gy / 8;                          	// die Page und die Pixelposition in Y-Richtung ermitteln
pos_y  = gy % 8;                          // nachfolgendes gilt fr einen Font mit Hhe von 6 Pixel
                                          // wenn pos_y = 7, dann steht der Text auf der Page/Zeile
                                          // wenn pos_y = 6, dannn ist der Text um ein Pixel nach oben verschoben
tmp = pos_y;                              // wenn pos_y < 6, dann muss der Text auf zwei Pages aufgeteilt werden

while(tmp < 7)                             // die Pixel solange nach oben schieben
   {
   for (i = 0; i < SMALL_FONT_WIDTH; i++)
		{
		upper_pix[i] >>= 1;						// das upper Bitmap oberhalb schieben
		if (lower_pix[i] & 0x01) upper_pix[i] |= 0x80; // wenn Bit.0 = 1, dann nach upper_ix retten
		lower_pix[i] >>= 1;						// das lower Bitmap schieben
		}
   tmp++;
   }

if (pos_y < 6)                            // dann mssen zwei Pages geschrieben werden
   {                                      // hier wird die obere Page geschreiben
   if (page)                            	// nur, wenn noch nicht die oberste Page errreicht ist
      {
      page--;                           	// die Page ber der Position anwhlen
      tmp = gx;                           // die Position merken
      for (i = 0; i < SMALL_FONT_WIDTH; i++)
         {
         if (tmp >= DOG_GX_PIXEL) break;  // beenden, wenn der Grafikbereich verlassen wird

         if (mode == SET_PIX)             // Pixel setzen
            {
            screen[page][tmp++] |= (upper_pix[i]);
            }

         else if (mode == CLR_PIX)         // Pixel lschen
              {
            screen[page][tmp++] &= ~(upper_pix[i]);
            }

         else                             // XOR Pixel
            {
            screen[page][tmp++] ^= (upper_pix[i]);
            }

         }
      page++;                           // zurck zur unteren Page
      }
   }

for (i = 0; i < SMALL_FONT_WIDTH; i++)    // den unteren Teil des Zeichens schreiben
   {
   if (gx >= DOG_GX_PIXEL) return;        // beenden, wenn der Grafikbereich verlassen wird


   if (mode == SET_PIX)             		// Pixel setzen
      {
      screen[page][gx++] |= (lower_pix[i]);
      }

   else if (mode == CLR_PIX)         		// Pixel lschen
      {
      screen[page][gx++] &= ~(lower_pix[i]);
      }

   else                             		// XOR Pixel
      {
      screen[page][gx++] ^= (lower_pix[i]);
      }
   }
}
#endif

// --- [ eof ] ----------------------------------------------------------------
